CON

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  LtServo   = 0               ' Servo for left wheel of BOE-BOT
  RtServo   = 1               ' Servo for right wheel of BOE-BOE
  PingServo = 2               ' Servo for PING bracket
  Ping      = 3               ' PING control line
  IRBuddy   = 4               ' IRBuddy control line
  DinDout   = 5               ' HM55B Din / Dout
  Clock     = 6               ' HM55B Clock
  NotEna    = 7               ' HM55B Not Enable
  TVBase    = 12              ' Base pin of 4 pin block for TV
  LEDmsb    = 16              ' MSB of 8 pin block for LEDs
  LEDlsb    = 23              ' LSB of 8 pin block for LEDs
  KeyData   = 26              ' PS/2 Keyboard Data pin
  KeyClk    = 27              ' PS/2 Keyboard Clock pin

  ' HM55B Commands
  RESET     = %0000           ' HM55B Reset command
  MEASURE   = %1000           ' HM55B Start measurement command
  REPORT    = %1100           ' HM55B Report measurement command
  READY     = %1100           ' HM55B Ready status
  
  XYBits    = 11              ' HM55B Number of bits for coordinate
  XYShft    = 32 - XYBits     ' HM55B Number of bits to shift

  ClrScr    = $00
  Home      = $01
  Bsp       = $08
  Tab       = $09
  SetX      = $0A
  SetY      = $0B
  SetClr    = $0C
  Cr        = $0D

  To_CM     = 29_034          ' Ping Time to Centimeters
  TVChannel = 2               ' Use tv broadcast output on channel 3
  IRMode    = %1100           ' Open mode, non-inverted, ignore tx echo on rx
  IRBaud    = 9600            ' Baud for IRBuddy (2400, 4800, or 9600)

OBJ
  text : "tv_text"
  ser : "FullDuplexSerial"
  
VAR
  long Servo[3], ServoStack[20], AtanInfo
  byte Buddy[8], Codes[7]

PUB start | w, i, s, a, x, y
  text.start(TVBase+TVChannel)
  text.str(string("Test BOE-BOT... V7"))
  startServo
  outa[IRBuddy]~
  dira[IRBuddy]~~                     ' Reset IRBuddy
  pause(-200)                         ' wait for 5ms
  dira[IRBuddy]~
  pause(-20)                          ' wait for 50ms
  ser.start(IRBuddy,IRBuddy,IRMode,IRBaud)
  w := 540
  s := 10
  repeat
    text.str(string(SetY,1,Cr,"Servo = "))
    text.dec(w)
    text.str(string("   "))
    SetServo(PingServo,w)
    text.str(string(SetY,2,Cr,"Angle = "))
    a := CompassValue
    x := a << 16 ~> 16
    y := a ~> 16
    text.hex(a,8)
    text.str(string("  "))
    text.dec(FindAtan(y,x) * 360 / 8192)
    text.str(string("   ",SetY,3,Cr,"Ping = "))
    text.dec(PingTicks(Ping) * 1000 / To_CM)
    text.str(string("   "))
    ser.str(string($72,5))
    repeat i from 0 to 7
      Buddy[i] := ser.rxtime(50)
    repeat i from 0 to 7
      if Buddy[i] < 254
        Codes[6] := Codes[5]
        Codes[5] := Codes[4]
        Codes[4] := Codes[3]
        Codes[3] := Codes[2]
        Codes[2] := Codes[1]
        Codes[1] := Codes[0]
        Codes[0] := Buddy[i]
    text.str(string(SetY,4,Cr,"IR = "))
    repeat i from 0 to 6
      text.out(SetX)
      text.out(i*4+5)
      text.dec(Codes[i])
      text.str(string("   "))
    if (w == 540) and (s < 0)
      s := 10
    if (w == 2320) and (s > 0)
      s := -10
    w += s
    pause(-20)

PRI CompassValue | x, y, p, status
    outa[NotEna..DinDout] := %100
    dira[NotEna..DinDout] := %110
    Toggle(NotEna)                                     ' make sure disabled briefly
    shiftout(DinDout, Clock, RESET, 4)
    Toggle(NotEna)                                     ' make sure disabled briefly
    shiftout(DinDout, Clock, MEASURE, 4)
    status := 0
    repeat until status == READY
      Toggle(NotEna)                                   ' briefly disable, then reenable
      shiftout(DinDout, Clock, REPORT, 4)
      status := shiftin(DinDout, Clock, 4)
    x := shiftin(DinDout, Clock, XYBits) << XYShft ~> XYShft
    y := shiftin(DinDout, Clock, XYBits) << XYShft ~> XYShft
    outa[NotEna] := 1                                  ' leave compass disabled
    return y << 16 | (x & $FFFF)

PRI pause(t)
  if t < 0                                             ' pause t seconds if t > 0
    t := clkfreq / ||t
  else                                                 ' pause 1/-t seconds if t < 0
    t *= clkfreq
  waitcnt(t + cnt)

PRI Toggle(pin)                                       
    outa[pin]~~                                        ' Put out brief pulse on NotEnable
    waitcnt(1000 + cnt)
    outa[pin]~

PRI ShiftIn(Dpin, Cpin, Bits) : Value | InBit
    dira[Dpin]~                                        ' Set data pin to input
    outa[Cpin]~                                        ' Set clock low 
    dira[Cpin]~~                                       ' Set clock pin to output 
    Value:=0                                                      
    REPEAT Bits                                        ' for number of bits                    
      outa[Cpin]~~                                     ' cycle clock                         
      outa[Cpin]~                                     
      InBit:= ina[Dpin]                                ' get bit value                          
      Value := (Value << 1) + InBit                    ' Add to value shifted by position                                         
      waitcnt(1000 + cnt)                              ' time delay

PRI ShiftOut(Dpin, Cpin, Value, Bits)
    outa[Dpin]~                                        ' Data pin = 0
    dira[Dpin]~~                                       ' Set data as output
    outa[Cpin]~
    dira[Cpin]~~                                       ' Set clock as output
    REPEAT Bits                                                                
      outa[Dpin] := Value >> (bits-1)                  ' Set output           
      Value := Value << 1                              ' Shift value right    
      outa[Cpin]~~                                     ' cycle clock          
      outa[Cpin]~                                                             
      waitcnt(1000 + cnt)                              ' delay                
    dira[Dpin]~                                        ' Set data as input

PRI PingTicks(Pin) | cnt1, cnt2, us
'' Return Ping)))'s one-way ultrasonic travel time in microseconds
    us := clkfreq / 1_000_000                                                                                 
    outa[Pin]~                                         ' Clear I/O Pin
    dira[Pin]~~                                        ' Make Pin Output
    outa[Pin]~~                                        ' Set I/O Pin
    outa[Pin]~                                         ' Clear I/O Pin (> 2 s pulse)
    dira[Pin]~                                         ' Make I/O Pin Input
    waitpne(0, |< Pin, 0)                              ' Wait For Pin To Go HIGH
    cnt1 := cnt                                        ' Store Current Counter Value
    waitpeq(0, |< Pin, 0)                              ' Wait For Pin To Go LOW 
    cnt2 := cnt                                        ' Store New Counter Value
    return (||(cnt1 - cnt2) / us) >> 1                 ' Return Time in s

PRI startServo : okay
  Servo[0] := 0                              ' Initialize pulse
  Servo[1] := 0                              '   width table
  Servo[2] := 0
  okay := cognew(DoServoPulses,@ServoStack)  ' Start servo pulse COG

PRI DoServoPulses | nextTime, pin, frame
  outa[LtServo..PingServo]~                  ' Initialize servo pins
  dira[LtServo..PingServo]~~
  frame := clkfreq / 50
  nextTime := cnt                            ' Every 20ms frame
  repeat
    repeat pin from LtServo to PingServo
      if Servo[pin] > 0                      ' If servo active
        outa[pin]~~
        waitcnt(Servo[pin] + cnt)            ' Output specified pulse
        outa[pin]~
    waitcnt(nextTime += frame)               ' Wait for next frame

PRI SetServo(pin,width) | us
  us := clkfreq / 1_000_000
  Servo[pin] := (width #> 500 <# 2500) * us  ' Set pulse width in us
  
PRI StopServo(pin)
  Servo[pin] := 0                            ' Stop sending pulses

PRI FindAtan(y,x) | delay
  AtanInfo := $FFC00000 | ((y & $7FF) << 11) | (x & $7FF)
  cognew(@getatan,@AtanInfo)
  delay := (clkfreq / 10) + cnt              ' Wait at most 100ms
  repeat while (AtanInfo < 0) AND ((delay - cnt) > 0)
  return AtanInfo

DAT
'
' Get arctangent of -ay/ax by solving for aa in: 0 = ax*sin(aa)+ay*cos(aa)
' Essentially search the sin/cos table for the smallest angle that best fits.
'
' on entry: PAR has the address of a long in HUB memory consisting of ax in the least
'           significant 11 bits and ay in the next 11 bits.  The rest of the long is
'           one bits so another cog can tell when it is replaced by the answer.
' on exit:  The long in HUB memory at PAR is replaced by the angle (from 0 to $1FFF)
'
getatan         rdlong   ax,PAR
                mov      ay,ax                  'get the values from the HUB
                shl      ax,#21
                sar      ax,#21                 'sign extend the 11 bit x value
                shl      ay,#10
                sar      ay,#21                 'sign extend the 11 bit y value
                mov      aa,lastaa
                mov      minaa,maxloop          'initialize loop variables
                mov      minval,maxint
                mov      loops,maxloop
:loop           mov      xx,ax
                mov      sin,aa
                call     #getsin                'compute ax * sin(aa)
                mov      yy,sin
                call     #multsign              'binary point between bits 16..15
                mov      term,yy
                mov      xx,ay
                mov      sin,aa
                call     #getcos                'compute ay * cos(aa)
                mov      yy,sin
                call     #multsign              'binary point between bits 16..15
                add      term,yy
                abs      term,term              'now we have ax*sin(aa)+ay*cos(aa)
                cmps     term,minval     wc     ' is ||term < minval?
         if_c   mov      minval,term
         if_c   mov      minaa,aa               'if so, save current angle and ||term
                sub      aa,#1
                djnz     loops,#:loop           'try next value
                wrlong   minaa,PAR
                cogid    term                   'save result and stop
                cogstop  term

ax              long     0                      'x value
ay              long     0                      'y value
aa              long     0                      'current angle
minaa           long     0                      'current angle for minimum
minval          long     0                      'current minimum
term            long     0                      'temporary
loops           long     0                      'loop counter
maxint          long     $7FFFFFFF              'largest possible positive long
maxloop         long     $2000                  'number of loops (# of angles)
lastaa          long     $1FFF                  'last angle in 4th quadrant

'
' Get sine / cosine
'
'      quadrant:  1             2             3             4
'         angle:  $0000..$07FF  $0800..$0FFF  $1000..$17FF  $1800..$1FFF
'   table index:  $0000..$07FF  $0800..$0001  $0000..$07FF  $0800..$0001
'        mirror:  +offset       -offset       +offset       -offset
'          flip:  +sample       +sample       -sample       -sample
'
' on entry: sin(12..0) holds angle (0 to just under 360)
' on exit:  sin holds signed value ranging from $0000FFFF ('1') to $FFFF0001 ('-1')
'
getcos          add     sin,sin_90              'for cosine, add 90
getsin          test    sin,sin_90      wc      'get quadrant 2|4 into c
                test    sin,sin_180     wz      'get quadrant 3|4 into nz
                negc    sin,sin                 'if quadrant 2|4, negate offset
                or      sin,sin_table           'or in sin table address >> 1
                shl     sin,#1                  'shift left to get final word address
                rdword  sin,sin                 'read word sample from $E000 to $F000
                negz    sin,sin                 'if quadrant 3|4, negate sample
getsin_ret
getcos_ret      ret                             '39..54 clocks
                                                '(variance is due to HUB sync on RDWORD)                                                
sin_90          long    $0800
sin_180         long    $1000
sin_table       long    $E000 >> 1              'sine table base shifted right
sin             long    0

'
' Multiply signed xx by yy.  In other words, xx(15..0) by yy(15,,0)
' [yy(31..16) must be sign bits on entry].  On exit, product in yy[31..0]
'
multsign        mov     ss,xx                   'compute the sign of the result
                xor     ss,yy
                abs     xx,xx                   'get the absolute value of the
                abs     yy,yy                   ' two factors
                shl     xx,#16                  'get multiplicand into x(31..16)
                mov     tt,#16                  'ready for 16 multiplier bits
                shr     yy,#1    wc             'get initial multiplier bit into c
:loop    if_c   add     yy,xx    wc             'if c set, add multiplicand into product
                rcr     yy,#1    wc             'get next multiplier bit into c, shift product
                djnz    tt,#:loop               'loop until done
                mov     ss,ss    wc             'set the sign of the product
                negnc   yy,yy
multsign_ret    ret

ss              long    0
xx              long    0
yy              long    0
tt              long    0